@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2005-2007, Scientific Computing Associates, Inc.
+# Copyright (c) 2005-2008, REvolution Computing, Inc.
 #
 # NetWorkSpaces is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License as published
@@ -63,6 +63,7 @@ __all__ = ['NwsServer', 'NetWorkSpace', 'NwsValueIterator',
            'NwsConnectException', 'NwsConnectionDroppedException',
            'NwsUnsupportedOperationException', 'NwsOperationException',
            'NwsNoWorkSpaceException', 'NwsDeclarationFailedException',
+           'NwsPicklingException', 'NwsUnpicklingException',
            'FIFO', 'LIFO', 'MULTI', 'SINGLE',
            'STRING', 'DICT',
            'WS_NAME', 'WS_MINE', 'WS_OWNER', 'WS_PERSISTENT', 'WS_NUMVARS', 'WS_VARLIST',
@@ -135,6 +136,12 @@ def _sanityFileObj(fobj, mode):
         else:
             raise TypeError('fobj must be a binary mode file object opened for reading')
 
+def _sanityStrObj(s, sname):
+    try:
+        assert len(s) == len(s.__str__())
+    except (AssertionError, AttributeError):
+        raise TypeError('%s must be a string' % sname)
+
 class NwsException(Exception):
     """Base class for all exceptions raised by this module."""
 
@@ -159,6 +166,37 @@ class NwsNoWorkSpaceException(NwsOperationException):
 class NwsDeclarationFailedException(NwsOperationException):
     """Variable declaration failed."""
 
+class NwsPicklingException(NwsOperationException):
+    """Error pickling an object."""
+
+    def __init__(self, exception, wsName, varName):
+        Exception.__init__(self)
+        self.exception = exception
+        self._wsName = wsName
+        self._varName = varName
+        try: self._exname = exception.__class__.__name__
+        except: self._exname = '[unknown]'
+
+    def __str__(self):
+        return "caught %s while pickling object to store in '%s @ %s'" % \
+                (repr(self._exname), self._wsName, self._varName)
+
+class NwsUnpicklingException(NwsOperationException):
+    """Error unpickling an object."""
+
+    def __init__(self, exception, wsName, varName, obj):
+        Exception.__init__(self)
+        self.exception = exception
+        self._wsName = wsName
+        self._varName = varName
+        self.obj = obj
+        try: self._exname = exception.__class__.__name__
+        except: self._exname = '[unknown]'
+
+    def __str__(self):
+        return "caught %s while unpickling object retrieved from '%s @ %s'" % \
+                (repr(self._exname), self._wsName, self._varName)
+
 class NwsServer(object):
 
     """Perform operations against the NWS server.
@@ -170,8 +208,6 @@ class NwsServer(object):
     def __init__(self, host='localhost', port=NWS_PORT):
         """Create a connection to the NWS server.
 
-        This constructor is intended for internal use only.
-
         Arguments:
 
         host -- Host name of the NWS server.
@@ -181,12 +217,16 @@ class NwsServer(object):
         """
         self.serverHost = host
         self.serverPort = port
+        self._connect()
+
+    def _connect(self):
         self.nwsSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
 
         try:
-            self.nwsSocket.connect((host, port))
+            self.nwsSocket.connect((self.serverHost, self.serverPort))
         except socket.error:
-            raise NwsConnectException('unable to connect to the NWS server at %s:%d' % (host, port))
+            raise NwsConnectException('unable to connect to the NWS server at %s:%d' % \
+                    (self.serverHost, self.serverPort))
 
         # denagle the socket to improve performance
         self.nwsSocket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
@@ -205,6 +245,13 @@ class NwsServer(object):
         self._handshake = self._recvN(4)
         self._cookieProtocol = self._handshake != _OldProtocol
 
+    def __getstate__(self):
+        return {'serverHost': self.serverHost, 'serverPort': self.serverPort}
+
+    def __setstate__(self, d):
+        self.__dict__.update(d)
+        self._connect()
+
     def __str__(self):
         return "NwsServer %s:%d" % (self.serverHost, self.serverPort)
 
@@ -278,8 +325,7 @@ class NwsServer(object):
         wsName -- Name of the workspace to delete.
 
         """
-        if type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        _sanityStrObj(wsName, 'workspace name')
 
         op = 'delete ws'
 
@@ -300,8 +346,7 @@ class NwsServer(object):
         not.
 
         """
-        if type(format) != str:
-            raise TypeError('format must be a string')
+        _sanityStrObj(format, 'format')
 
         if format not in _formats:
             raise ValueError('illegal format: ' + format)
@@ -366,8 +411,7 @@ class NwsServer(object):
                 >>> workspace = server.openWs(name)
 
         """
-        if type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        _sanityStrObj(wsName, 'workspace name')
 
         op = 'mktemp ws'
         self._sendAll('0002%020d%s%020d%s' % (len(op), op, len(wsName), wsName))
@@ -442,8 +486,7 @@ class NwsServer(object):
                     "__init__() got an unexpected keyword argument '%s'" % \
                     unrecog[0])
 
-        if type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        _sanityStrObj(wsName, 'workspace name')
 
         # if invoked directly by user, we need to create a space
         # instance. if invoked via NetWorkSpace constructor, use the
@@ -530,8 +573,7 @@ class NwsServer(object):
                     "__init__() got an unexpected keyword argument '%s'" % \
                     unrecog[0])
 
-        if type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        _sanityStrObj(wsName, 'workspace name')
 
         # if invoked directly by user, we need to create a space
         # instance. if invoked via networkspace constructor, use the
@@ -651,10 +693,11 @@ class NetWorkSpace(object):
                     "__init__() got an unexpected keyword argument '%s'" % \
                     unrecog[0])
 
-        if type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        _sanityStrObj(wsName, 'workspace name')
 
         self.curWs = wsName
+        self._useUse = useUse
+        self._opt = opt
 
         # if invoked (indirectly) via a NwsServer openWs or useWs method,
         # the server will be passed in and used. if invoked directly,
@@ -663,10 +706,10 @@ class NetWorkSpace(object):
             self.server = NwsServer(serverHost, serverPort)
             # now give the server a chance to do its thing.
             try:
-                if useUse:
-                    self.server.useWs(wsName, self, **opt)
+                if self._useUse:
+                    self.server.useWs(self.curWs, self, **self._opt)
                 else:
-                    self.server.openWs(wsName, self, **opt)
+                    self.server.openWs(self.curWs, self, **self._opt)
             except Exception, e:
                 # close the server and re-raise the exception
                 try: self.server.close()
@@ -680,6 +723,29 @@ class NetWorkSpace(object):
         self._handshake = self.server._handshake
         self._cookieProtocol = self.server._cookieProtocol
 
+    def __getstate__(self):
+        return {'curWs': self.curWs, '_useUse': self._useUse, '_opt': self._opt,
+                'server': self.server}
+
+    def __setstate__(self, d):
+        self.__dict__.update(d)
+
+        try:
+            if self._useUse:
+                self.server.useWs(self.curWs, self, **self._opt)
+            else:
+                self.server.openWs(self.curWs, self, **self._opt)
+        except Exception, e:
+            # close the server and re-raise the exception
+            try: self.server.close()
+            except: pass
+            raise e
+
+        self.send = self.server._sendAll
+        self.recv = self.server._recvN
+        self._handshake = self.server._handshake
+        self._cookieProtocol = self.server._cookieProtocol
+
     def __str__(self):
         return "NetWorkspace '%s' [%s]" % (self.curWs, str(self.server))
 
@@ -725,13 +791,11 @@ class NetWorkSpace(object):
         mode -- Mode of the variable.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
+        _sanityStrObj(mode, 'variable mode')
 
-        if type(mode) != str:
-            raise TypeError('mode must be a string')
-
-        if mode not in _modes:
+        # note that workspace variable extensions start with '__'
+        if mode not in _modes and not mode.startswith('__'):
             raise ValueError('illegal mode: ' + str(mode))
 
         op = 'declare var'
@@ -753,8 +817,7 @@ class NetWorkSpace(object):
         varName -- Name of the workspace variable to delete.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         op = 'delete var'
 
@@ -764,6 +827,7 @@ class NetWorkSpace(object):
         if status: raise NwsOperationException('deleteVar failed')
 
     def __retrieve(self, varName, op, missing=None, fobj=None):
+        _sanityStrObj(varName, 'variable name')
         _sanityFileObj(fobj, 'w')
 
         self.send('0003%020d%s%020d%s%020d%s' % \
@@ -784,7 +848,11 @@ class NetWorkSpace(object):
             if desc & _DirectString:
                 return pVal
             elif pVal:
-                return cPickle.loads(pVal)
+                try:
+                    val = cPickle.loads(pVal)
+                except Exception, e:
+                    raise NwsUnpicklingException(e, self.curWs, varName, pVal)
+                return val
             else:
                 return missing
         else:
@@ -810,9 +878,6 @@ class NetWorkSpace(object):
         varName -- Name of the workspace variable to fetch.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         return self.__retrieve(varName, 'fetch', None)
 
     def fetchTry(self, varName, missing=None):
@@ -836,11 +901,10 @@ class NetWorkSpace(object):
         missing -- Value to return if the variable has no values.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         try:
             return self.__retrieve(varName, 'fetchTry', missing)
+        except NwsUnpicklingException:
+            raise
         except NwsOperationException:
             return missing
 
@@ -861,9 +925,6 @@ class NetWorkSpace(object):
         varName -- Name of the workspace variable to find.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         return self.__retrieve(varName, 'find', None)
 
     def findTry(self, varName, missing=None):
@@ -886,11 +947,10 @@ class NetWorkSpace(object):
                 default is None.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         try:
             return self.__retrieve(varName, 'findTry', missing)
+        except NwsUnpicklingException:
+            raise
         except NwsOperationException:
             return missing
 
@@ -915,11 +975,10 @@ class NetWorkSpace(object):
                 is 'string'.
 
         """
-        if wsName is not None and type(wsName) != str:
-            raise TypeError('workspace name must be a string')
+        if wsName is not None:
+            _sanityStrObj(wsName, 'workspace name')
 
-        if type(format) != str:
-            raise TypeError('format must be a string')
+        _sanityStrObj(format, 'format')
 
         if format not in _formats:
             raise ValueError('illegal format: ' + format)
@@ -961,19 +1020,22 @@ class NetWorkSpace(object):
         val -- Value to store in the variable.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         op = 'store'
 
         desc = _PythonFP
-        if StringType == type(val): desc |= _DirectString
+        if isinstance(val, str):
+            desc |= _DirectString
         descTxt = '%020u' % desc
 
         if desc & _DirectString:
             pVal = val
         else:
-            pVal = cPickle.dumps(val)
+            try:
+                pVal = cPickle.dumps(val, 1)
+            except Exception, e:
+                raise NwsPicklingException(e, self.curWs, varName)
 
         self.send('0005%020d%s%020d%s%020d%s%020d%s%020d' % \
                 (len(op), op, len(self.curWs), self.curWs,
@@ -995,9 +1057,6 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         return self.__retrieve(varName, 'fetch', fobj=fobj)
 
     def fetchTryFile(self, varName, fobj):
@@ -1012,11 +1071,10 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         try:
             return self.__retrieve(varName, 'fetchTry', fobj=fobj)
+        except NwsUnpicklingException:
+            raise
         except NwsOperationException:
             return False
 
@@ -1032,9 +1090,6 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         return self.__retrieve(varName, 'find', fobj=fobj)
 
     def findTryFile(self, varName, fobj):
@@ -1049,11 +1104,10 @@ class NetWorkSpace(object):
         fobj -- File to write data to.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
         try:
             return self.__retrieve(varName, 'findTry', fobj=fobj)
+        except NwsUnpicklingException:
+            raise
         except NwsOperationException:
             return False
 
@@ -1077,9 +1131,7 @@ class NetWorkSpace(object):
                 file.  The default value is zero.
 
         """
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
-
+        _sanityStrObj(varName, 'variable name')
         _sanityFileObj(fobj, 'r')
 
         op = 'store'
@@ -1171,7 +1223,11 @@ class NetWorkSpace(object):
             if desc & _DirectString:
                 return (status, pVal, varId, valIndex, n)
             elif pVal:
-                return (status, cPickle.loads(pVal), varId, valIndex, n)
+                try:
+                    val = cPickle.loads(pVal)
+                except Exception, e:
+                    raise NwsUnpicklingException(e, self.curWs, varName, pVal)
+                return (status, val, varId, valIndex, n)
             else:
                 raise StopIteration
         else:
@@ -1200,8 +1256,7 @@ class NetWorkSpace(object):
         if not self.server._cookieProtocol:
             raise NwsUnsupportedOperationException('NWS server does not support ifetch: ' + self._handshake)
 
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         return NwsValueIterator(self, varName, 'ifetch')
 
@@ -1225,8 +1280,7 @@ class NetWorkSpace(object):
         if not self.server._cookieProtocol:
             raise NwsUnsupportedOperationException('NWS server does not support ifetchTry: ' + self._handshake)
 
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         return NwsValueIterator(self, varName, 'ifetchTry')
 
@@ -1252,8 +1306,7 @@ class NetWorkSpace(object):
         if not self.server._cookieProtocol:
             raise NwsUnsupportedOperationException('NWS server does not support ifind: ' + self._handshake)
 
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         return NwsValueIterator(self, varName, 'ifind')
 
@@ -1278,11 +1331,18 @@ class NetWorkSpace(object):
         if not self.server._cookieProtocol:
             raise NwsUnsupportedOperationException('NWS server does not support ifindTry: ' + self._handshake)
 
-        if type(varName) != str:
-            raise TypeError('variable name must be a string')
+        _sanityStrObj(varName, 'variable name')
 
         return NwsValueIterator(self, varName, 'ifindTry')
 
+    def makefile(self, varName, mode='r'):
+        if mode in ['r', 'w', 'a']:
+            return NwsTextFile(self, varName, mode)
+        elif mode in ['rb', 'wb', 'ab']:
+            raise ValueError('mode %r is not supported yet' % mode)
+        else:
+            raise ValueError('illegal mode specified: %r' % mode)
+
 class NwsValueIterator(object):
 
     """Implements the iterated operations against a workspace variable.
@@ -1381,3 +1441,160 @@ class NwsValueIterator(object):
         except StopIteration, e:
             self._stopped = True
             return 0
+
+class NwsTextFile(object):
+    def __init__(self, ws, varName, mode):
+        assert mode in ['r', 'w', 'a']
+        self._ws = ws
+        self._varName = varName
+        self._wbuffer = []
+        self._rbuffer = ''
+        self.closed = False
+        self.mode = mode
+        self.name = varName
+        self.softspace = False
+
+        if mode == 'w':
+            # delete the variable if it exists
+            try: ws.deleteVar(varName)
+            except NwsOperationException: pass
+
+        if mode == 'r':
+            self._it = ws.ifindTry(varName)
+
+        ws.declare(varName, FIFO)
+
+    def __del__(self):
+        try: self.close()
+        except: pass
+
+    def __iter__(self):
+        if self.mode not in ['r', 'rb']:
+            raise ValueError('iteration not allowed in mode %r' % self.mode)
+        return self
+
+    def __repr__(self):
+        state = self.closed and "closed" or "open"
+        return "<%s nws file %r, mode %r at %s>" % (state, self.name, self.mode, hex(id(self)))
+
+    def __str__(self):
+        return repr(self)
+
+    def next(self):
+        x = self.readline()
+        if not x:
+            raise StopIteration()
+        return x
+
+    def read(self, n=-1):
+        if self.closed:
+            raise ValueError('I/O operation on closed file')
+
+        if self.mode not in ['r', 'rb']:
+            raise ValueError('read operation not allowed in mode %r' % self.mode)
+
+        navail = len(self._rbuffer)
+
+        # read one or more values from the nws variable until
+        # we have enough data, or we run out of values
+        if n < 0 or navail < n:
+            buffer = [self._rbuffer]
+            try:
+                while n < 0 or navail < n:
+                    x = self._it.next()
+                    if not isinstance(x, str):
+                        raise ValueError('read a non-string from %r: %r' % (self, x))
+                    buffer.append(x)
+                    navail += len(x)
+            except StopIteration:
+                pass
+            self._rbuffer = ''.join(buffer)
+
+        if n < 0:
+            x = self._rbuffer
+            self._rbuffer = ''
+        else:
+            x = self._rbuffer[:n]
+            self._rbuffer = self._rbuffer[n:]
+        return x
+
+    def readline(self):
+        if self.closed:
+            raise ValueError('I/O operation on closed file')
+
+        if self.mode not in ['r', 'rb']:
+            raise ValueError('read operation not allowed in mode %r' % self.mode)
+
+        x = self._rbuffer
+        buffer = [x]
+        try:
+            while not x.endswith('\n'):
+                x = self._it.next()
+                if not isinstance(x, str):
+                    raise ValueError('read a non-string from %r: %r' % (self, x))
+                buffer.append(x)
+        except StopIteration:
+            pass
+
+        x = ''.join(buffer)
+        self._rbuffer = ''
+        return x
+
+    def readlines(self):
+        return [x for x in self]
+
+    def write(self, s):
+        if self.closed:
+            raise ValueError('I/O operation on closed file')
+
+        if self.mode not in ['w', 'a', 'wb', 'ab']:
+            raise ValueError('write operation not allowed in mode %r' % self.mode)
+
+        if s:
+            x = s.split('\n')
+            if len(x) > 1:
+                self._wbuffer.append(x[0])
+                y = ''.join(self._wbuffer)
+                self._ws.store(self._varName, y + '\n')
+                for y in x[1:-1]:
+                    self._ws.store(self._varName, y + '\n')
+                self._wbuffer = []
+                if x[-1]:
+                    self._wbuffer.append(x[-1])
+            else:
+                self._wbuffer.append(s)
+
+    def writelines(self, lst):
+        for line in lst:
+            self.write(line)
+
+    def close(self):
+        if not self.closed:
+            try:
+                if self.mode in ['w', 'wb', 'a', 'ab']:
+                    self.flush()
+            finally:
+                try: del self._wbuffer
+                except AttributeError: pass
+                try: del self._rbuffer
+                except AttributeError: pass
+                self.closed = True
+
+    def flush(self):
+        if self.closed:
+            raise ValueError('I/O operation on closed file')
+
+        if self.mode not in ['w', 'a', 'wb', 'ab']:
+            raise ValueError('flush operation not allowed in mode %r' % self.mode)
+
+        # flush any buffered data without a newline
+        if self._wbuffer:
+            y = ''.join(self._wbuffer)
+            if y:
+                self._ws.store(self._varName, y)
+            self._wbuffer = []
+
+    def isatty(self):
+        if self.closed:
+            raise ValueError('I/O operation on closed file')
+        return False
