@@ -1,5 +1,5 @@
 #
-# Copyright (c) 2005-2007, Scientific Computing Associates, Inc.
+# Copyright (c) 2005-2008, REvolution Computing, Inc.
 #
 # NetWorkSpaces is free software; you can redistribute it and/or
 # modify it under the terms of the GNU General Public License as published
@@ -92,7 +92,7 @@ __all__ = [
     'SleighGatheredException', 'SleighStoppedException',
     'SleighOccupiedException', 'SleighIllegalValueException',
     'SleighScriptException', 'SleighTaskException',
-    'SleighJoinException'
+    'SleighJoinException', 'SleighNotAllowedException'
 ]
 
 # We use alternating barriers to synchronize eachWorker
@@ -202,6 +202,9 @@ class SleighTaskException(SleighException):
 class SleighJoinException(SleighException):
     """Too late to join worker group."""
 
+class SleighNotAllowedException(SleighException):
+    """Sleigh construction not allowed."""
+
 # functions needed for pickling and unpickling code objects
 def _code_ctor(*args):
     return new.code(*args)
@@ -270,8 +273,15 @@ def cmdLaunch(verbose=0):
 
     PythonSleighWorkerCount -- The total number of workers.
 
+    PythonSleighRedirectOutput -- Specifies whether stdout and stderr
+            should be redirected to a workspace variable.
+            The default value is "True".
+
     """
 
+    # note that xbool works quite differently than bool.
+    # xbool('False') -> False, whereas bool('False') -> True
+    xbool = lambda s: s.lower().strip() == 'true'
     args = {
         'PythonSleighNwsName': ('nwsName', str),
         'PythonSleighUserNwsName': ('userNwsName', str),
@@ -279,20 +289,22 @@ def cmdLaunch(verbose=0):
         'PythonSleighNwsPort': ('nwsPort', int),
         'PythonSleighWorkerCount': ('maxWorkerCount', int),
         'PythonSleighName': ('name', str),
+        'PythonSleighRedirectOutput': ('redirectOutput', xbool),
     }
     kw = {'verbose': verbose}
     for e, a in args.items():
         try: kw[a[0]] = a[1](_Env[e])
         except KeyError: pass
 
+    print >> sys.stderr, "Calling launch:", kw
     launch(**kw)
 
 def launch(nwsName, nwsHost='localhost', nwsPort=8765,
         maxWorkerCount=-1, name=socket.getfqdn(), verbose=0,
-        userNwsName='__default'):
+        userNwsName='__default', redirectOutput=True):
     """Execute sleigh tasks on behalf of the sleigh owner.
 
-    launch(nwsName[, nwsHost, nwsPort, maxWorkerCount, name, verbose, userNwsName])
+    launch(nwsName[, nwsHost, nwsPort, maxWorkerCount, name, verbose, userNwsName, redirectOutput])
 
     This function is executed from the sleigh worker script, via
     cmdLaunch, but it can also be executed manually, using a command
@@ -324,6 +336,9 @@ def launch(nwsName, nwsHost='localhost', nwsPort=8765,
     userNwsName -- Name of the user workspace.  The default value is
             '__default'.
 
+    redirectOutput -- Boolean flag that indicates if stdout and stderr
+            should be redirected to a workspace variable.
+
     """
     nws = NetWorkSpace(nwsName, nwsHost, nwsPort, useUse=True, create=False)
     userNws = nws.server.useWs(userNwsName, create=False)
@@ -375,7 +390,7 @@ def launch(nwsName, nwsHost='localhost', nwsPort=8765,
     workerCount = nws.find('workerCount')
 
     # enter the main worker loop
-    _workerLoop(nws, displayName, rank, workerCount, verbose, userNws)
+    _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectOutput)
 
 class Logger(object):
     def __init__(self, ws, name, role):
@@ -389,7 +404,7 @@ class Logger(object):
         except Exception, e:
             msg = 'Exception while logging: %s' % str(e)
 
-        print >> sys.stderr, msg
+        print >> sys.__stderr__, msg
         if self._ws:
             try:
                 logmsg = '[%s] %s %s -- %s' % \
@@ -405,7 +420,7 @@ class Logger(object):
     def logDebug(self, *s):
         self._logMsg(*s, **{'var': 'logDebug'})
 
-def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
+def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws, redirectOutput):
     """Repeatedly execute tasks on behalf of the sleigh.
 
     _workerLoop(nws, displayName, rank, workerCount, verbose, userNws)
@@ -426,6 +441,9 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
 
     userNws -- NetWorkSpace object for the user workspace.
 
+    redirectOutput -- Boolean flag that indicates if stdout and stderr
+            should be redirected to a workspace variable.
+
     """
     bx = 0
 
@@ -445,6 +463,16 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
 
     logger = Logger(nws, displayName, 'worker')
 
+    if redirectOutput:
+        f = nws.makefile('output@%d' % rank, 'w')
+        sys.stdout = f
+        sys.stderr = f
+
+    # try to prevent Sleigh objects from being accidentally
+    # created when modules are imported
+    if not _Env.has_key('PythonSleighAllowed'):
+        _Env['PythonSleighAllowed'] = 'FALSE'
+
     try:
         while 1:
             # update the number of tasks executed
@@ -506,8 +534,24 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
 
                         # get the function from the module
                         f = getattr(module, t.funcName)
+                    except ImportError, e:
+                        r.value = 'Unable to import module %s that defines worker ' \
+                                  'function %s: %s' % \
+                            (repr(t.module), repr(t.funcName), str(e))
+                        r.error = True
+                        logger.logError(r.value)
+                    except AttributeError, e:
+                        r.value = 'Unable to find worker function %s at ' \
+                                  'top level of module %s: %s' % \
+                            (repr(t.funcName), repr(t.module), str(e))
+                        r.error = True
+                        logger.logError(r.value)
                     except Exception, e:
-                        r.value = 'Task definition from module failed: ' + str(e)
+                        try: exname = e.__class__.__name__
+                        except: exname = '[unknown]'
+                        r.value = 'Error getting worker function %s from ' \
+                                  'module %s - %s: %s' % \
+                            (repr(t.funcName), repr(t.module), exname, str(e))
                         r.error = True
                         logger.logError(r.value)
 
@@ -530,11 +574,17 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
                         except Exception, e:
                             tb = sys.exc_info()[2]
                             fname, lineno = _tb_info(tb)
-                            r.value = 'Task invocation failed: %s [%s:%d]' % \
-                                    (str(e), fname, lineno)
+                            try: exname = e.__class__.__name__
+                            except: exname = '[unknown]'
+                            r.value = 'Task invocation failed - %s: %s [%s:%d]' % \
+                                    (repr(exname), str(e), fname, lineno)
                             r.error = True
                             logger.logError(r.value)
                             if verbose:
+                                # check if there is an attribute that provides
+                                # the "originating" exception
+                                if hasattr(e, 'exception'):
+                                    logger.logError(str(e.exception))
                                 logger.logError(_tb_stack(tb))
             elif t.type == 'eval':
                 try:
@@ -589,7 +639,8 @@ def _workerLoop(nws, displayName, rank, workerCount, verbose, userNws):
                 nws.find(_barrierNames[bx])
                 bx ^= 1
     except Exception, e:
-        logger.logError('Worker exiting: ' + str(e))
+        # connection may corrupt, so just write to stderr
+        sys.__stderr__.write('Worker exiting: %s\n' % str(e))
         # traceback.print_exc()
 
 def getparam(name, opts, key, defkey='default', defval=None, coerce=str):
@@ -910,6 +961,7 @@ defaultSleighOptions = {
     'wsNameTemplate':   'sleigh_ride_%04d',
     'userWsNameTemplate': 'sleigh_user_%04d',
     'verbose':          False,
+    'redirectOutput':   True,
 }
 
 class Accumulator(object):
@@ -1131,6 +1183,12 @@ class Sleigh(object):
                 working directory (as controled by the workerDir
                 argument).  The default value is False.
 
+        redirectOutput -- Boolean flag used to control whether messages
+                written by the worker functions to stdout and stderr
+                should be redirected to a workspace variable.  If False,
+                they will go to the output file if verbose is True,
+                otherwise to the null device.  The default value is True.
+
         """
         # sanity check the keyword arguments
         unrecog = [a for a in kw.keys() if a not in defaultSleighOptions.keys()]
@@ -1139,6 +1197,12 @@ class Sleigh(object):
                     "__init__() got an unexpected keyword argument '%s'" % \
                     unrecog[0])
 
+        # check for a recursive sleigh construction
+        if _Env.get('PythonSleighAllowed', 'TRUE') == 'FALSE':
+            raise SleighNotAllowedException(
+                    'Sleigh construction not allowed '
+                    '(probably accidentally called from sleigh worker)')
+
         # join related
         self.rankCount = 0
 
@@ -1360,6 +1424,7 @@ class Sleigh(object):
         envVars.append('PythonSleighName=' + machine)
         envVars.append('PythonSleighWorkerCount=' + str(self.workerCount))
         envVars.append('PythonSleighID=' + str(id))
+        envVars.append('PythonSleighRedirectOutput=' + str(self.options['redirectOutput']))
 
         # remote execution command
         if self.options['launch'] == 'local':
